import argparse
import getpass
import sys
from base64 import b32decode
from collections import OrderedDict, defaultdict, namedtuple
from functools import reduce

import totp

_subcommands = {}

_argument = namedtuple('argument', ['args', 'kwargs'])
def argument(*args, **kwargs):
    return _argument(args, kwargs)

def subcommand(name, *args, **kwargs):
    def decorator(func):
        _subcommands[name] = (args, kwargs, func)
    return decorator

def _parse_args(args):
    parser = argparse.ArgumentParser(
        description='Print a TOTP token getting the shared key from pass(1).'
    )

    subparsers = parser.add_subparsers(dest='command')
    aliases = {}

    for name, (_args, kwargs, func) in _subcommands.items():
        _parser = subparsers.add_parser(name, **kwargs)
        _parser.set_defaults(func=func)
        for arg in _args:
            _parser.add_argument(*arg.args, **arg.kwargs)

        _aliases = kwargs.get('aliases', ())
        for alias in _aliases:
            aliases[alias] = name

    def replace_aliases(args, aliases):
        for i, arg in enumerate(args):
            if arg in aliases:
                args[i] = aliases[arg]

    def add_default_subcommand_if_omitted(args, default):
        if not any(arg in ('-h', '--help') or arg in subparsers.choices for arg in args):
            args.insert(0, default)

    replace_aliases(args, aliases)
    add_default_subcommand_if_omitted(args, 'show')

    return parser.parse_args(args)

def run():
    args = _parse_args(sys.argv[1:])
    try:
        args.func(args)
    except totp.BackendError as e:
        print('%s returned an error:\n%s' % (e.backend_name, e))
        raise SystemExit(-1)
    except KeyboardInterrupt:
        print()

@subcommand(
    'add',
    argument(
        'identifier',
        help='the identifier under the \'2fa\' folder where the key should be saved'),
    argument(
        '-u',
        '--uri',
        help='an optional otpauth uri to read the entry data from'),
    aliases=['-a'],
    description='Add a new TOTP entry to the database.',
    help='add a new TOTP entry to the database')
def _cmd_add(args):
    if args.uri:
        add_uri(args.identifier, args.uri)
    else:
        add_interactive(args.identifier)

def input_shared_key():
    while True:
        try:
            shared_key = totp.normalize_secret(getpass.getpass('Shared key: '))
            b32decode(shared_key.upper())
            if shared_key == "":
                raise ValueError('The key entered was empty')
            return shared_key
        except ValueError as err:
            print(*err.args)

def add_interactive(path):
    token_length = input('Token length [%d]: ' % totp.DIGITS_DEFAULT)
    token_length = int(token_length) if token_length else totp.DIGITS_DEFAULT

    shared_key = input_shared_key()

    totp.add_pass_entry(path, token_length, shared_key)

def add_uri(path, uri):
    totp.add_pass_entry_from_uri(path, uri)

@subcommand(
    'show',
    argument(
        '-s',
        dest='offset_seconds',
        metavar='SECONDS',
        default=0,
        help='offset the clock by the given number of seconds'),
    argument(
        '-n', '--nocopy',
        action="store_true",
        help='Do not copy the token, only show it.'),
    argument(
        'identifier',
        help='the identifier by which the key can be found under the \'2fa\' folder'),
    description='Show the current TOTP token for a registered entry.',
    help='(default action) show the current TOTP token for a registered entry')
def _cmd_show(args):
    totp.generate_token(args.identifier, seconds=args.offset_seconds, to_clipboard=not args.nocopy)

if __name__ == '__main__':
    run()



#!/usr/bin/env python
#
# Print a TOTP token getting the shared key from pass(1).

import os
import platform
import re
import subprocess
import sys
import urllib.parse

import onetimepass


DIGITS_DEFAULT = 6


class BackendError(Exception):
    backend_name = '<none>'

class PassBackendError(BackendError):
    backend_name = 'pass'

    def __init__(self, err):
        if isinstance(err, bytes):
            err = err.decode('utf-8', 'replace')
        super().__init__(err.rstrip('\n'))

class ValidationError(Exception): pass

def validate(*args):
    for (validator, error_message) in args:
        if not validator():
            raise ValidationError(error_message)


def get_length(pass_entry):
    """Return the required token length."""
    for line in pass_entry:
        if line.lower().startswith('digits:'):
            return int(re.search(r'\d+', line).group())

    return 6


def add_pass_entry_from_uri(path, uri):
    parsed = parse_otpauth_uri(uri)
    add_pass_entry(path, parsed['digits'], parsed['secret'])


def add_pass_entry(path, token_length, shared_key):
    """Add a new entry via pass."""
    code_path = "2fa/{}/code"
    code_path = code_path.format(path)

    pass_entry = "{}\ndigits: {}\n".format(shared_key, token_length)

    p = subprocess.Popen(
        ['pass', 'insert', '-m', '-f', code_path],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    pass_output, err = p.communicate(
        input=bytearray(pass_entry, encoding='utf-8')
    )

    if len(err) > 0:
        raise PassBackendError(err)


def get_pass_entry(path):
    """Return the entire entry as provided via pass."""
    code_path = "2fa/{}/code"
    code_path = code_path.format(path)

    p = subprocess.Popen(
        ['pass', code_path],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )

    pass_output, err = p.communicate()

    if len(err) > 0:
        raise PassBackendError(err)

    return pass_output.decode()


def copy_to_clipboard(text):
    try:
        if platform.system() == 'Darwin':
            command = ['pbcopy']
        elif platform.system() == 'Windows':
            command = ['clip']
        else:
            selection = os.environ.get(
                'PASSWORD_STORE_X_SELECTION',
                'clipboard',
            )
            command = ['xclip', '-selection', selection]

        p = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        p.stdin.write(text)
        p.stdin.close()
        p.wait()
    except FileNotFoundError:
        print(
            '{} not found. Not copying code'.format(command[0]),
            file=sys.stderr,
        )


def normalize_secret(secret):
    s = secret.replace(' ', '')

    if len(s) % 8 != 0:
        num_needed_padding_chars = 8 - (len(s) % 8)
        s += '=' * num_needed_padding_chars

    return s


def generate_token(path, seconds=0, to_clipboard=True):
    """Generate the TOTP token for the given path and the given time offset"""
    import time
    clock = time.time() + float(seconds)

    pass_entry = get_pass_entry(path)

    # Remove the trailing newline or any other custom data users might have
    # saved:
    pass_entry = pass_entry.splitlines()
    secret = normalize_secret(pass_entry[0])

    digits = get_length(pass_entry)
    token = onetimepass.get_totp(secret, as_string=True, token_length=digits,
                                 clock=clock)

    print(token.decode())

    if to_clipboard:
        copy_to_clipboard(token)


def parse_otpauth_uri(uri):
    parsed = urllib.parse.urlsplit(uri)
    query = urllib.parse.parse_qs(parsed.query)

    secret = query.get('secret', [])
    digits = query.get('digits', [])
    issuer = query.get('issuer', [])

    validate(
        (lambda: parsed.scheme == 'otpauth', 'invalid URI scheme: %s' % parsed.scheme),
        (lambda: parsed.netloc == 'totp', 'unsupported key type: %s' % parsed.netloc),
        (lambda: len(secret) <= 1, 'too many \'secret\' arguments'),
        (lambda: len(digits) <= 1, 'too many \'digits\' arguments'),
        (lambda: len(issuer) <= 1, 'too many \'issuer\' arguments'),
        (lambda: len(secret) == 1, 'no secret found'),
    )

    secret, = secret
    issuer = issuer[0] if issuer else None
    digits = int(digits[0]) if digits else DIGITS_DEFAULT

    return {
        'secret': secret,
        'digits': digits,
        'issuer': issuer,
        'label': parsed.path.lstrip('/'),
    }



#!/usr/bin/env python3

import sys

from setuptools import setup

if sys.version_info < (3, 3):
    raise RuntimeError(
        'totp-cli requires python >= 3.3, but this python is %d.%d' %
        sys.version_info[0:2]
    )

setup(
    name='totp',
    description='A cli based TOTP app.',
    long_description=open('README.rst').read(),
    author='Hugo Osvaldo Barrera',
    author_email='hugo@barrera.io',
    url='https://github.com/WhyNotHugo/totp-cli/',
    license='ISC',
    packages=['totp'],
    entry_points={
        'console_scripts': [
            'totp = totp.cli:run',
        ]
    },
    install_requires=[
        open('requirements.txt').readlines()
    ],
    use_scm_version={
        'version_scheme': 'post-release',
        # 'write_to': 'totp/version.py',
    },
    setup_requires=['setuptools_scm'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Console :: Curses',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Topic :: Utilities',
    ]
)
